from typing_extensions import AsyncIterable, TypeVar, Generic, ParamSpec, \
  Awaitable, Callable, Coroutine, Any
from functools import wraps, cached_property
from datetime import timedelta, datetime
from pydantic import RootModel, TypeAdapter
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlalchemy.exc import DatabaseError
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy.sql.expression import func
from sqltypes import ValidatedJSON
from sqlmodel import select, text, or_
from sqlmodel.ext.asyncio.session import AsyncSession
from pipeteer.queues import Queue, ListQueue, QueueError, InfraError, InexistentItem
from pipeteer.util import type_arg

T = TypeVar('T')
U = TypeVar('U')
Ps = ParamSpec('Ps')

SessionFn = Callable[[AsyncSession], U]

def wrap_exceptions(fn: Callable[Ps, Coroutine[Any, Any, T]]) -> Callable[Ps, Coroutine[Any, Any, T]]:
  @wraps(fn)
  async def wrapped(*args: Ps.args, **kwargs: Ps.kwargs) -> T:
    try:
      return await fn(*args, **kwargs)
    except DatabaseError as e:
      raise InfraError(e) from e
  
  return wrapped

exec_options = {}
# exec_options = {'isolation_level': 'SERIALIZABLE', 'no_cache': True}

class SqlQueue(Queue[T], Generic[T]):

  @staticmethod
  def new(type: type[U], url: str, *, table: str, echo: bool = False) -> 'SqlQueue[U]':
    engine = create_async_engine(url, echo=echo)
    return SqlQueue(type, engine, table=table)

  def __init__(self, type: type[T], engine: AsyncEngine, *, table: str):
    self.engine = engine
    self.table = table
    self.session: AsyncSession | None = None
    self.type = type

    class Base(DeclarativeBase):
      ...

    class Table(Base):
      __tablename__ = table
      key: Mapped[str] = mapped_column(primary_key=True)
      value: Mapped[RootModel[Any]] = mapped_column(type_=ValidatedJSON(type, name='JSON')) # type: ignore
      ttl: Mapped[datetime|None] = mapped_column(default=None)

    self.Table = Table
    self.metadata = Base.metadata
    self.initialized = False

  async def initialize(self):
    if not self.initialized:
      try:
        async with self.engine.begin() as conn:
          await conn.run_sync(self.metadata.create_all)
          if self.engine.dialect.name == 'sqlite':
            await conn.execute(text("PRAGMA journal_mode=WAL"))
            await conn.execute(text("PRAGMA busy_timeout=200"))
        self.initialized = True
      except DatabaseError:
        ...

  def __repr__(self):
    return f'SqlQueue(engine={self.engine!r}, table={self.Table.__tablename__!r})'
  
  async def with_session(self, f: SessionFn[Awaitable[U]]) -> U:
    """Generates a session on-the-fly if executing without a transaction"""
    try:
      await self.initialize()
      if self.session is None:
        async with AsyncSession(self.engine) as s:
          return await f(s)
      else:
        return await f(self.session)
    except DatabaseError as e:
      raise InfraError(e) from e
  
  async def with_autocommit(self, f: SessionFn[Awaitable[U]]) -> U:
    """Generates a session on-the-fly if executing without a transaction. Autocommits at the end"""
    try:
      await self.initialize()
      if self.session is None:
        async with AsyncSession(self.engine) as s:
          out = await f(s)
          await s.commit()
          return out
      else:
        return await f(self.session)
    except DatabaseError as e:
      raise InfraError(e) from e
  

  async def push(self, key: str, value: T):
    async def _push(s: AsyncSession):
      stmt = select(self.Table).where(self.Table.key == key)
      row = (await s.exec(stmt)).first()
      if row is not None:
        await s.delete(row)
      s.add(self.Table(key=key, value=value, ttl=datetime.now()))
    
    return await self.with_autocommit(_push)
    

  async def pop(self, key: str):
    async def _pop(s: AsyncSession):
      stmt = select(self.Table).where(self.Table.key == key)
      row = (await s.exec(stmt)).first()
      if row is None:
        raise InexistentItem(key)
      
      await s.delete(row)

    return await self.with_autocommit(_pop)


  async def read(self, key: str, /, *, reserve: timedelta | None = None) -> T:
    async def _read(s: AsyncSession) -> T:
      stmt = select(self.Table).where(
        self.Table.key == key, 
        or_(self.Table.ttl < datetime.now(), self.Table.ttl == None)
      )
      if reserve is not None:
        stmt = stmt.with_for_update(nowait=True, skip_locked=True)

      row = (await s.exec(stmt, execution_options=exec_options if reserve else {})).first()
      if row:
        if reserve is not None:
          row.ttl = datetime.now() + reserve
          s.add(row)
          await s.commit()
        return row.value # type: ignore
      
      raise InexistentItem(key)
  
    return await self.with_session(_read)
  
  
  async def read_any(self, *, reserve: timedelta | None = None) -> tuple[str, T] | None:
    async def _read_any(s: AsyncSession) -> tuple[str, T] | None:
      stmt = select(self.Table).where(
        or_(self.Table.ttl < datetime.now(), self.Table.ttl == None)
      ).limit(1).order_by(func.random())
      if reserve is not None:
        stmt = stmt.with_for_update(skip_locked=True)
      
      row = (await s.exec(stmt, execution_options=exec_options if reserve else {})).first()
      if row:
        k, v = row.key, row.value
        if reserve is not None:
          row.ttl = datetime.now() + reserve
          s.add(row)
          await s.commit()
        
        return k, v # type: ignore
      
    return await self.with_session(_read_any)
    
  async def has(self, key: str, /, *, reserve: timedelta | None = None) -> bool:
    async def _has(s: AsyncSession) -> bool:
      stmt = select(self.Table).where(
        self.Table.key == key, 
        or_(self.Table.ttl < datetime.now(), self.Table.ttl == None)
      )
      if reserve is not None:
        stmt = stmt.with_for_update(nowait=True, skip_locked=True)

      row = (await s.exec(stmt, execution_options=exec_options if reserve else {})).first()
      if row and reserve is not None:
        row.ttl = datetime.now() + reserve
        s.add(row)
        await s.commit()
      
      return row is not None
    
    return await self.with_session(_has)

    
  async def items(self, *, reserve: timedelta | None = None, max: int | None = None) -> AsyncIterable[tuple[str, T]]: 
    await self.initialize()
    try:
      async with AsyncSession(self.engine) as s:
        stmt = select(self.Table).where(
          or_(self.Table.ttl < datetime.now(), self.Table.ttl == None)
        ).limit(max)
        if reserve is not None:
          stmt = stmt.with_for_update(skip_locked=True)

        result = await s.exec(stmt, execution_options=exec_options if reserve else {})
        for row in result:
          k, v = row.key, row.value
          if reserve is not None:
            row.ttl = datetime.now() + reserve
            s.add(row)
            
          yield k, v # type: ignore
        
        if reserve is not None:
          await s.commit()
          
    except DatabaseError as e:
      raise InfraError(e) from e
    
  async def _clear(self, s: AsyncSession):
    await s.execute(text(f'DELETE FROM "{self.table}"'))
    
  async def clear(self):
    return await self.with_autocommit(self._clear)
  
  @wrap_exceptions
  async def enter(self, other=None):
    if isinstance(other, SqlQueue) and other.engine.url == self.engine.url:
      self.session = other.session
    elif self.session is None:
      self.session = await AsyncSession(self.engine).__aenter__()

  @wrap_exceptions
  async def commit(self, other=None):
    if not self.session:
      raise QueueError('No transaction to commit')
    
    if not isinstance(other, SqlQueue) or other.engine.url != self.engine.url:
      await self.session.commit()

  @wrap_exceptions
  async def close(self, other=None):
    if self.session and (not isinstance(other, SqlQueue) or other.engine.url != self.engine.url):
      await self.session.close()
    
  @wrap_exceptions
  async def rollback(self, other=None):
    if not self.session:
      raise QueueError('No transaction to rollback')
    
    if not isinstance(other, SqlQueue) or other.engine.url != self.engine.url:
      await self.session.rollback()
    
class ListSqlQueue(ListQueue[T], SqlQueue[list[T]], Generic[T]):

  @cached_property
  def adapter(self):
    return TypeAdapter(self.type)
  
  @cached_property
  def single_adapter(self):
    return TypeAdapter(type_arg(self.type))

  async def append(self, key: str, value: T):
    async def _append(s: AsyncSession):
      single = self.adapter.dump_json([value]).decode()
      obj = self.single_adapter.dump_json(value).decode()
      if s.bind.dialect.name == 'postgresql':
        stmt = f'''
          INSERT INTO "{self.table}" (key, value)
            VALUES (:key, jsonb(:single))
            ON CONFLICT (key)
            DO UPDATE SET 
              value = "{self.table}".value || jsonb(:obj)
        '''
      else:
        stmt = f'''
          INSERT INTO "{self.table}" (key, value)
            VALUES (:key, json(:single))
            ON CONFLICT(key)
            DO UPDATE SET 
              value = json_insert(value, '$[#]', json(:obj))
        '''
      stmt = text(stmt).bindparams(key=key, single=single, obj=obj)
      await s.execute(stmt)

    return await self.with_autocommit(_append)
    


from typing_extensions import TypeVar, Generic, Callable, Awaitable
import asyncio
from dataclasses import dataclass
from pipeteer import WriteQueue

A = TypeVar('A')
B = TypeVar('B')

class tee(WriteQueue[A], Generic[A]):
  """A queue that pushes to the multiple queues at once"""

  def __init__(self, q1: WriteQueue[A], q2: WriteQueue[A], /, *qs: WriteQueue[A], ordered: bool = False):
    self.queues: list[WriteQueue[A]] = [q1, q2, *qs]
    self.ordered = ordered

  def __repr__(self) -> str:
    reprs = ', '.join(repr(q) for q in self.queues)
    return f'tee({reprs})'
  
  async def push(self, key: str, value: A):
    if self.ordered:
      for q in self.queues:
        await q.push(key, value)
    else:
      await asyncio.gather(*[q.push(key, value) for q in self.queues])


@dataclass
class prefilter(WriteQueue[A], Generic[A]):
  queue: WriteQueue[A]
  pred: Callable[[tuple[str, A]], Awaitable[bool]]

  def __repr__(self):
    return f'prefilter({self.queue!r})'

  async def push(self, key: str, value: A):
    if await self.pred((key, value)):
      return await self.queue.push(key, value)

@dataclass
class premap(WriteQueue[B], Generic[A, B]):

  queue: WriteQueue[A]
  mapper: Callable[[str, B], Awaitable[tuple[str, A]]]

  def __repr__(self):
    return f'premap({self.queue!r})'

  async def push(self, key: str, value: B):
    k, v = await self.mapper(key, value)
    return await self.queue.push(k, v)


from .queues import ReadQueue, WriteQueue, Queue, ListQueue, Transaction, Transactional
from .backend import Backend
from .pipelines import Pipeline, Runnable, Inputtable, Observable, Context, \
  activity, workflow, WorkflowContext, task, multitask, Client

__all__ = [
  'ReadQueue', 'WriteQueue', 'Queue', 'ListQueue', 'Transaction', 'Transactional',
  'Backend', 'Pipeline', 'Runnable', 'Inputtable', 'Observable',
  'Context', 'activity', 'workflow', 'WorkflowContext',
  'task', 'multitask', 'Client',
]


from typing_extensions import TypeVar, Generic, Callable, Any
from dataclasses import dataclass
from pipeteer.pipelines import Pipeline, Context
from pipeteer.queues import Queue, ReadQueue, WriteQueue, Routed, RoutedQueue, ops
from pipeteer.util import param_type, type_arg, num_params, Func2or3

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
Ctx = TypeVar('Ctx', bound=Context)
Artifact = TypeVar('Artifact')

@dataclass
class Task(Pipeline[A, B, Ctx, Artifact], Generic[A, B, Ctx, Artifact]):
  call: Callable[[ReadQueue[A], WriteQueue[B], Ctx], Artifact]

  def Qurls(self, ctx: Ctx) -> Queue[str]:
    return ctx.backend.queue(self.id+'-urls', str)
  
  def Qin(self, ctx: Ctx) -> Queue[A]:
    return ctx.backend.queue(self.id, self.Tin)

  def input(self, ctx: Ctx) -> WriteQueue[Routed[A]]:
    return ops.tee(
      self.Qurls(ctx).premap(lambda x: x['url']),
      self.Qin(ctx).premap(lambda x: x['value'])
    )
  
  def observe(self, ctx: Ctx):
    return { 'input': self.Qin(ctx), 'urls': self.Qurls(ctx) }
  
  def run(self, ctx: Ctx, /):
    Qin = self.Qin(ctx)
    Qout = RoutedQueue(self.Qurls(ctx), lambda url: ctx.backend.queue_at(url, self.Tout))
    return self.call(Qin, Qout, ctx)

def task(id: str | None = None):
  def decorator(fn: Func2or3[ReadQueue[A], WriteQueue[B], Ctx, Artifact]) -> Task[A, B, Ctx, Artifact]:
    Tin = type_arg(param_type(fn, 0) or Any) # type: ignore
    if Tin is None:
      raise TypeError(f'Task {fn.__name__} must have a type hint for its input type')
    
    Tout = type_arg(param_type(fn, 1) or Any) # type: ignore
    if Tout is None:
      raise TypeError(f'Task {fn.__name__} must have a type hint for its output type')
    
    return Task(
      id=id or fn.__name__,
      Tin=Tin, Tout=Tout,
      call=fn if num_params(fn) == 3 else (lambda Qin, Qout, _: fn(Qin, Qout)) # type: ignore
    )
      
  return decorator


from typing_extensions import TypeVar
from urllib.parse import urlparse, parse_qs, unquote
from pydantic import BaseModel
from pipeteer.queues import Queue, http

T = TypeVar('T')

class SqlParams(BaseModel):
  table: str

class HttpParams(BaseModel):
  secret: str | None = None

def parse(url: str, type: type[T]) -> Queue[T]:

  parsed_url = urlparse(url) # 'file://path/to/base?prefix=hello'
  scheme = parsed_url.scheme # 'file'
  netloc = parsed_url.netloc # 'path'
  path = unquote(parsed_url.path) # '/to/base'
  endpoint = netloc + path # 'path/to/base'
  query = parse_qs(parsed_url.query) # { 'prefix': ['hello'] }
  query = { k: v[0] for k, v in query.items() }

  if scheme in ('http', 'https'):
    params = HttpParams(**query)
    url = f'{scheme}://{endpoint}'
    return http.QueueClient(url, type=type, secret=params.secret)
  
  elif scheme.startswith('sql+'):
    params = SqlParams(**query)
    from pipeteer.queues import SqlQueue
    proto = scheme.removeprefix('sql+')
    return SqlQueue.new(type, url=f'{proto}://{endpoint}', table=params.table)
  
  else:
    raise ValueError(f'Unknown scheme: {scheme}')


from typing_extensions import TypeVar, Any
from dataclasses import dataclass, field
from fastapi import FastAPI, Request
from pipeteer.queues import http, Queue, Routed
from pipeteer import Backend, Runnable, Inputtable, Observable, Context

A = TypeVar('A')
Ctx = TypeVar('Ctx', bound=Context)
AnyT: type = Any # type: ignore

@dataclass
class HttpBackend(Backend):
  base_url: str
  token: str | None = None
  app: FastAPI = field(default_factory=FastAPI)
  id2urls: dict[str, str] = field(default_factory=dict)
  urls2id: dict[str, str] = field(default_factory=dict)
  queues: dict[str, Queue] = field(default_factory=dict)
  pipelines: set[str] = field(default_factory=set)
  mounted: bool = False

  def __post_init__(self):
    if self.token is not None:
      @self.app.middleware('http')
      async def check_token(request: Request, call_next):
        if request.headers.get('Authorization') != f'Bearer {self.token}':
          return 'Unauthorized', 401
        return await call_next(request)

  @property
  def url(self) -> str:
    return self.base_url.rstrip('/')

  def public_queue(self, id: str, type: type[A]) -> tuple[str, Queue[A]]:
    queue = self.queue(id, type)
    if not id in self.id2urls:
      self.app.mount(f'/callbacks/{id}', http.queue_api(queue, type))
      url = f'{self.url}/callbacks/{id}'
      self.id2urls[id] = url
      self.urls2id[url] = id

    return self.id2urls[id], queue
  
  def queue_at(self, url: str, type: type[A]) -> Queue[A]:
    if url in self.urls2id:
      id = self.urls2id[url]
      return self.queue(id, type)
    else:
      return Queue.of(url, type)

  
  def mount(self, pipeline: Runnable[Any, Any, Ctx, Any] | Inputtable[Any, Any, Ctx], ctx: Ctx):
    
    @self.app.get(f'/pipelines')
    def list_pipelines():
      return {
        id: f'{self.url}/pipelines/{id}'
        for id in self.pipelines
      }
    
    self.pipelines.add(pipeline.id)

    urls = {}
    
    if isinstance(pipeline, Observable):
      urls['queues'] = f'{self.url}/pipelines/{pipeline.id}/queues'
      queues = pipeline.observe(ctx)
      for name, queue in queues.items():
        self.app.mount(f'/pipelines/{pipeline.id}/queues/{name}', http.queue_api(queue, AnyT))

      @self.app.get(f'/pipelines/{pipeline.id}/queues')
      def observe():
        return {
          name: f'{self.url}/pipelines/{pipeline.id}/queues/{name}'
          for name in queues
        }

    if isinstance(pipeline, Inputtable):
      urls['input'] = f'{self.url}/pipelines/{pipeline.id}/input/write'
      self.app.mount(f'/pipelines/{pipeline.id}/input/write', http.write_api(pipeline.input(ctx), Routed[pipeline.Tin]))

    @self.app.get('/pipelines/{id}')
    def list_pipeline(id: str):
      return urls
    


from typing_extensions import TypeVar, Generic, get_args
from dataclasses import dataclass
from pipeteer.queues import WriteQueue, ListQueue

A = TypeVar('A')

@dataclass
class singleton(WriteQueue[A], Generic[A]):
  queue: ListQueue[A]

  async def push(self, key: str, val: A):
    await self.queue.push(key, [val])

@dataclass
class appender(WriteQueue[A], Generic[A]):
  queue: ListQueue[A]

  async def push(self, key: str, val: A):
    await self.queue.append(key, val)


from datetime import timedelta
from typing_extensions import AsyncIterable, TypeVar, Generic, ParamSpec, Awaitable, Callable
from dataclasses import dataclass
from datetime import datetime, timedelta
import jwt
from pydantic import TypeAdapter, ValidationError
try:
  import httpx
except ImportError:
  raise ImportError('Please install `httpx` to use HTTP Queue clients')
from pipeteer.queues import Queue, ReadQueue, WriteQueue, QueueError, InfraError, Transactional

T = TypeVar('T')
U = TypeVar('U')
Ps = ParamSpec('Ps')

def sign_token(secret: str, expiry: datetime | None = None) -> str:
  payload = {} if expiry is None else {'exp': expiry.timestamp()}
  return jwt.encode(payload, secret, algorithm='HS256')

class ClientMixin(Transactional, Generic[T]):
  def __init__(
    self, url: str, *, type: 'type[T]',
    headers: dict[str, str] = {}, secret: str | None = None
  ):
    self.url = url
    self.type = type
    self.adapter = TypeAdapter(type)
    self.client: httpx.AsyncClient | None = None
    self.headers = headers
    self.secret = secret

  def params(self) -> dict[str, str] | None:
    if self.secret:
      return {'token': sign_token(self.secret, datetime.now() + timedelta(minutes=5))}
  
  async def enter(self, other=None):
    self.client = httpx.AsyncClient(headers=self.headers)
  
  async def close(self, other=None):
    if self.client:
      await self.client.aclose()
      self.client = None

  async def with_client(self, f: Callable[[httpx.AsyncClient], Awaitable[U]]) -> U:
    try:
      if self.client is None:
        async with httpx.AsyncClient(headers=self.headers, params=self.params()) as client:
          return await f(client)
      else:
        return await f(self.client)
    except httpx.RequestError as e:
      raise InfraError(e) from e
    
  async def with_client_iter(self, f: Callable[[httpx.AsyncClient], AsyncIterable[U]]) -> AsyncIterable[U]:
    try:
      if self.client is None:
        async with httpx.AsyncClient(headers=self.headers, params=self.params()) as client:
          async for item in f(client):
            yield item
      else:
        async for item in f(self.client):
          yield item
    except httpx.RequestError as e:
      raise InfraError(e) from e

def urljoin(*parts: str) -> str:
  return '/'.join(part.strip('/') for part in parts)

err_adapter = TypeAdapter(QueueError)
@dataclass
class WriteClient(ClientMixin[T], WriteQueue[T], Generic[T]):

  def __init__(self, url: str, type: 'type[T]', headers: dict[str, str] = {}, secret: str | None = None):
    super().__init__(url, type=type, headers=headers, secret=secret)
    self.dump = self.adapter.dump_json

  async def push(self, key: str, value: T):
    async def _push(client: httpx.AsyncClient):
      url = urljoin(self.url, 'write', key)
      r = await client.post(
        url,
        data=self.dump(value), # type: ignore
        headers=self.headers,
      )
      if r.status_code != 200:
        try:
          raise err_adapter.validate_json(r.content)
        except ValidationError:
          raise QueueError(f'Error pushing to {self.url}: {r.content}. Key: {key}, Value: {value}')
        
    return await self.with_client(_push)
    
@dataclass
class ReadClient(ClientMixin[T], ReadQueue[T], Generic[T]):

  def __init__(self, url: str, *, type: 'type[T]', headers: dict[str, str] = {}, secret: str | None = None):
    super().__init__(url, type=type, headers=headers, secret=secret)
    self.parse = self.adapter.validate_json
    self.parse_entry = TypeAdapter(tuple[str, self.type]|None).validate_json
  
  async def pop(self, key: str):
    async def _pop(client: httpx.AsyncClient):
      url = urljoin(self.url, 'read/item', key)
      r = await client.delete(url)
      if r.status_code != 200:
        try:
          raise err_adapter.validate_json(r.content)
        except ValidationError as e:
          raise QueueError(f'Error popping from {self.url}: {r.content}')
        
    return await self.with_client(_pop)
    
  
  async def read(self, key: str, /, *, reserve: timedelta | None = None) -> T:
    async def _read(client: httpx.AsyncClient):
      url = urljoin(self.url, 'read/item', key)
      params = {'reserve': reserve.total_seconds()} if reserve is not None else {}
      r = await client.get(url, params=params)
      if r.status_code != 200:
        try:
          raise err_adapter.validate_json(r.content)
        except ValidationError as e:
          raise QueueError(f'Error reading from {self.url}: {r.content}')
        
      return self.parse(r.content)
    
    return await self.with_client(_read)

  async def read_any(self, *, reserve: timedelta | None = None) -> tuple[str, T] | None:
    async def _ready_any(client: httpx.AsyncClient):
      url = urljoin(self.url, 'read/item')
      params = {'reserve': reserve.total_seconds()} if reserve is not None else {}
      r = await client.get(url, params=params)
      if r.status_code != 200:
        try:
          raise err_adapter.validate_json(r.content)
        except ValidationError as e:
          raise QueueError(f'Error reading from {self.url}: {r.content}')
        
      return self.parse_entry(r.content)
    
    return await self.with_client(_ready_any)
  
  def keys(self) -> AsyncIterable[str]:
    async def _keys(client: httpx.AsyncClient):
      url = urljoin(self.url, 'read/keys')
      r = await client.get(url)
      if r.status_code != 200:
        try:
          raise err_adapter.validate_json(r.content)
        except ValidationError as e:
          raise QueueError(f'Error reading keys from {self.url}: {r.content}')
        
      for key in TypeAdapter(list[str]).validate_json(r.content):
        yield key

    return self.with_client_iter(_keys)
    
  async def clear(self):
    async def _clear(client: httpx.AsyncClient):
      url = urljoin(self.url, 'read') + '/'
      r = await client.delete(url)
      if r.status_code != 200:
        try:
          raise err_adapter.validate_json(r.content)
        except ValidationError as e:
          raise QueueError(f'Error clearing {self.url}: {r.content}')
    
    return await self.with_client(_clear)

  def items(self, *, reserve: timedelta | None = None, max: int | None = None) -> AsyncIterable[tuple[str, T]]:
    return super().items(reserve=reserve, max=max)
    

class QueueClient(Queue[T], WriteClient[T], ReadClient[T], Generic[T]):
  def __init__(self, url: str, *, type: 'type[T]', headers: dict[str, str] = {}, secret: str | None = None):
    ReadClient.__init__(self, url, type=type)
    WriteClient.__init__(self, url, type, headers=headers, secret=secret)
    


from typing_extensions import Literal, Any
from dataclasses import dataclass

@dataclass(eq=False)
class QueueError(Exception):
  """Base class for all queue errors"""

  detail: Any | None = None
  key: str | None = None
  value: Any | None = None

  def __str__(self) -> str:
    return self.__repr__()
  
@dataclass
class InfraError(QueueError):
  """Infrastructure-related error (not logic)"""
  reason: Literal['infra-error'] = 'infra-error'

@dataclass(eq=False)
class InexistentItem(QueueError):
  reason: Literal['inexistent-item'] = 'inexistent-item'
  
ReadError = InfraError | InexistentItem


from typing_extensions import TypeVar
from dataclasses import dataclass
from multiprocessing import Process
from pipeteer.queues import Queue, ListQueue, ZeroMQueue, SqlQueue, ListSqlQueue
from pipeteer.backend.sql import SqlBackend
import asyncio

A = TypeVar('A')

@dataclass
class ZmqBackend(SqlBackend):

  def queue(self, id: str, type: type[A]) -> Queue[A]:
    queue = SqlQueue(type, self.engine, table=id)
    return ZeroMQueue.of(queue, topic=id, type=type)
  
  def list_queue(self, id: str, type: type[A]) -> ListQueue[A]:
    return ListSqlQueue(list[type], self.engine, table=id)
  
  def run(self):
    from pipeteer.queues.zeromq import proxy
    def _run():
      asyncio.run(proxy())
    return Process(target=_run)


from typing_extensions import TypeVar
from datetime import datetime
import jwt
from pydantic import TypeAdapter, ValidationError
from fastapi import FastAPI, Request, Response
from fastapi.exceptions import RequestValidationError
from pipeteer.queues import ReadQueue, WriteQueue, Queue, QueueError, InexistentItem

T = TypeVar('T')

def verify_token(*, token: str, secret: str, now: datetime | None = None) -> bool:
  now = now or datetime.now()
  try:
    exp = jwt.decode(token, secret, algorithms=['HS256'], options={'verify_exp': False}).get('exp')
    return exp is None or now < datetime.fromtimestamp(exp)
  except jwt.PyJWTError:
    return False
  
def token_middleware(secret: str):
  async def middleware(req: Request, call_next):
    token = req.query_params.get('token')
    if token and verify_token(token=token, secret=secret):
      return await call_next(req)
    else:
      return Response(status_code=401, content='Unauthorized')
  return middleware

def write_api(queue: WriteQueue[T], type: type[T], *, secret: str | None = None) -> FastAPI:
  app = FastAPI(generate_unique_id_function=lambda route: route.name)
  if secret:
    app.middleware('http')(token_middleware(secret))

  parse = TypeAdapter(type).validate_json

  @app.post('/{key:path}', responses={500: {'model': QueueError}, 404: {'model': InexistentItem}})
  async def push(key: str, req: Request, res: Response):
    try:
      value = parse(await req.body())
      await queue.push(key, value)
    except ValidationError as e:
      raise RequestValidationError(e.errors())
    except InexistentItem as e:
      res.status_code = 404
      return e
    except QueueError as e:
      res.status_code = 500
      return e
    
  return app


def read_api(queue: ReadQueue[T], *, secret: str | None = None) -> FastAPI:
  app = FastAPI(generate_unique_id_function=lambda route: route.name)
  if secret:
    app.middleware('http')(token_middleware(secret))

  @app.delete('/item/{key:path}', responses={500: {'model': QueueError}, 404: {'model': InexistentItem}})
  async def pop(key: str, r: Response):
    try:
      return await queue.pop(key)
    except InexistentItem as e:
      r.status_code = 404
      return e
    except QueueError as e:
      r.status_code = 500
      return e
    
  @app.get('/item', responses={500: {'model': QueueError}})
  async def read_any(r: Response) -> tuple[str, T] | None:
    try:
      return await queue.read_any()
    except QueueError as e:
      r.status_code = 500
      return e # type: ignore
    
  @app.get('/item/{key:path}', responses={500: {'model': QueueError}, 404: {'model': InexistentItem}})
  async def read(key: str, r: Response) -> T:
    try:
      return await queue.read(key)
    except InexistentItem as e:
      r.status_code = 404
      return e # type: ignore
    except QueueError as e:
      r.status_code = 500
      return e # type: ignore
    
  @app.get('/keys', responses={500: {'model': QueueError}})
  async def keys(r: Response) -> list[str]:
    try:
      return [key async for key in queue.keys()]
    except QueueError as e:
      r.status_code = 500
      return e # type: ignore
    
  @app.delete('/', responses={500: {'model': QueueError}})
  async def clear(r: Response):
    try:
      await queue.clear()
    except QueueError as e:
      r.status_code = 500
      return e # type: ignore

  return app


def queue_api(queue: Queue[T], type: type[T], *, secret: str | None = None) -> FastAPI:
  app = FastAPI(generate_unique_id_function=lambda route: route.name)
  app.mount('/write', write_api(queue, type))
  app.mount('/read', read_api(queue))
  if secret:
    app.middleware('http')(token_middleware(secret))
  return app


from .client import WriteClient, ReadClient, QueueClient
from .server import write_api, read_api, queue_api

__all__ = [
  'WriteClient', 'ReadClient', 'QueueClient',
  'write_api', 'read_api', 'queue_api',
]


from .types import num_params, param_type, return_type, type_arg, Func1or2, Func2or3
from .asyn import race, exp_backoff

__all__ = [
  'num_params', 'param_type', 'return_type', 'type_arg', 'Func1or2', 'Func2or3',
  'race', 'exp_backoff',
]


from datetime import timedelta
from typing import Any, Coroutine
from typing_extensions import AsyncIterable, Callable, TypeVar, Generic
from dataclasses import dataclass
from pipeteer.queues import ReadQueue, WriteQueue, Queue

T = TypeVar('T')

@dataclass
class ReadDelegate(ReadQueue[T], Generic[T]):
  """A queue that delegates operations to its `queue` attribute"""
  queue: ReadQueue[T]

  def read(self, key: str, /, *, reserve: timedelta | None = None):
    return self.queue.read(key, reserve=reserve)
  
  def read_any(self, *, reserve: timedelta | None = None):
    return self.queue.read_any(reserve=reserve)
  
  def pop(self, key: str):
    return self.queue.pop(key)

  def wait_any(self, *, reserve: timedelta | None = None, poll_interval: timedelta = timedelta(seconds=1)):
    return self.queue.wait_any(reserve=reserve, poll_interval=poll_interval)
  
  def safe_read(self, key: str, /, *, reserve: timedelta | None = None):
    return self.queue.safe_read(key, reserve=reserve)
  
  def items(self, *, reserve: timedelta | None = None, max: int | None = None):
    return self.queue.items(reserve=reserve, max=max)
  
  def keys(self):
    return self.queue.keys()
  
  def values(self):
    return self.queue.values()
  
  def has(self, key: str, /, *, reserve: timedelta | None = None):
    return self.queue.has(key, reserve=reserve)
  
  def clear(self):
    return self.queue.clear()



@dataclass
class WriteDelegate(WriteQueue[T], Generic[T]):
  queue: WriteQueue[T]

  def push(self, key: str, value: T):
    return self.queue.push(key, value)
  
  def premap(self, mapper):
    return self.queue.premap(mapper)
  
  def premap_k(self, mapper):
    return self.queue.premap_k(mapper)
  
  def premap_kv(self, mapper):
    return self.queue.premap_kv(mapper)
  

@dataclass
class Delegate(ReadDelegate[T], WriteDelegate[T], Generic[T]):
  queue: Queue[T] # type: ignore


from typing_extensions import TypeVar, Generic, Self, Mapping
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, replace, KW_ONLY
from dslog import Logger
from pipeteer.backend import Backend
from pipeteer.queues import Queue, WriteQueue, Routed


A = TypeVar('A')
B = TypeVar('B')
Artifact = TypeVar('Artifact', covariant=True)

@dataclass
class Context:
  backend: Backend
  _: KW_ONLY
  log: Logger = field(default_factory=lambda: Logger.click().limit('INFO'))

  def prefix(self, prefix: str) -> Self:
    return replace(self, log=self.log.prefix(prefix))

Ctx = TypeVar('Ctx', bound=Context)

@dataclass
class Inputtable(Generic[A, B, Ctx]):
  Tin: type[A]
  Tout: type[B]
  id: str

  def input(self, ctx: Ctx) -> WriteQueue[Routed[A]]:
    return ctx.backend.queue(self.id, Routed[self.Tin])
  
@dataclass
class Runnable(ABC, Generic[A, B, Ctx, Artifact]):
  Tin: type[A]
  Tout: type[B]
  id: str
  
  @abstractmethod
  def run(self, ctx: Ctx, /) -> Artifact:
    ...

class Observable(ABC, Generic[Ctx]):
  @abstractmethod
  def observe(self, ctx: Ctx) -> Mapping[str, Queue]:
    ...

class Pipeline(Runnable[A, B, Ctx, Artifact], Inputtable[A, B, Ctx], Observable[Ctx], Generic[A, B, Ctx, Artifact]):
  def as_client(self, url: str) -> 'Pipeline[A, B, Context, Artifact]':
    """Modify (in-place) the pipeline to act as `Client(url)`"""
    self.url = url
    self.input = lambda ctx: ctx.backend.queue_at(self.url, Routed[self.Tin])
    return self # type: ignore
